%%%%%%%%%%%%%%%%%%%%%%%%%%%%% -*- Mode: Prolog -*- %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% uni.pl --- 
%% Author          : Massimo Fasciano
%% Created On      : Sat Jul 11 01:47:54 1992
%% Last Modified By: Massimo Fasciano
%% Last Modified On: Mon Oct 18 17:43:54 1993
%% Update Count    : 84
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- ensure_loaded('include.pl').
:- ensure_loaded('morpho.pl').
:- ensure_loaded('tools.pl').
:- ensure_loaded('punif.pl').
:- ensure_loaded('unif.pl').

%%%%%%%% THE UNIFIER %%%%%%%%%%

% sub-FDs are in cset if they contain the "cat" feature.
% FUF also uses the "pattern" and "cset" features for this purpose,
% (easy to add to our system)...
in_cset(FD) :- nonvar(FD), FD=[cat:_|_], !.
in_cset(FD) :- nonvar(FD), FD=[_|R], in_cset(R).

% the main inference engine as described in the paper
% a) call the grammar
% b) check for recursive unifications (CSET)
% c) perform them
uni(FD) :- gram(FD),uni2(FD).
uni2(V) :- var(V),!.
uni2([Feature:FD|R]) :- in_cset(FD),!,uni(FD),uni2(R).
uni2([_|R]) :- uni2(R).

%%%%%%%% THE LINEARIZER %%%%%%%%%%

% unify and then linearize an FD
unilin(FD,L) :-
    uni(FD),lin(FD,L).

% linearize an FD using the "pattern" special features.
lin(FD,L) :-
    unif(FD,[pattern:Pats|_]),
    nonvar(Pats),!,
    pattern_list_unif(Pats,Pat),
    lin2(FD,Pat,[],L).
lin(FD,[FD]).

lin2(_,[],L,L).
lin2(FD,[H|T],Lin,Lout) :-
    unif(FD,[H:SubFD|_]),
    nonvar(SubFD),!,
    lin(SubFD,L),conc(Lin,L,Lin2),lin2(FD,T,Lin2,Lout).
lin2(FD,[...|T],Lin,Lout) :-
    !,conc(Lin,[...],Lin2),lin2(FD,T,Lin2,Lout).
lin2(FD,[H|T],Lin,Lout) :- lin2(FD,T,Lin,Lout).

pattern_list_unif([Pat],Pat).
pattern_list_unif([Pat1,Pat2|R],Res) :-
    punif(Pat1,Pat2,Pat),pattern_list_unif([Pat|R],Res).
